# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

from tkinter import messagebox, ttk

from customtkinter import CTkButton, CTkFrame, CTkLabel

from ice_launcher.components.heading import Heading
from ice_launcher.components.scrollable_frame import ScrollableFrame
from ice_launcher.lib import accounts, msa


class Accounts(CTkFrame):
    def __init__(self, master) -> None:
        super().__init__(master=master)

        self.grid_columnconfigure(0, weight=1)

        heading = Heading(self, "👥 Accounts")
        heading.grid(row=0, column=0, pady=20, padx=20, sticky="nwe")

        self.accounts_list = ScrollableFrame(master=self)
        self.accounts_list.grid(row=1, column=0, pady=20, padx=20, sticky="nswe")
        self.grid_rowconfigure(1, weight=1)

        button_bar = CTkFrame(master=self)
        button_bar.grid(row=100, column=0, pady=0, padx=0, sticky="nwe")

        add_account_button = CTkButton(
            master=button_bar,
            text="Add Account ✨",
            command=self.add_account,
        )
        add_account_button.grid(row=0, column=0, pady=10, padx=10, sticky="nse")
        button_bar.grid_columnconfigure(1, weight=1)

        self.update_accounts_list()

    def add_account(self) -> None:
        accounts.login_account(self.update_accounts_list)

    def select_account(self, account_id: str) -> None:
        accounts.set_active_account(account_id)
        self.update_accounts_list()

    def remove_account(self, account_id: str) -> None:
        if messagebox.askyesno(
            "Remove account", "Are you sure you want to remove this account?"
        ):
            accounts.remove_account(account_id)
            self.update_accounts_list()

    def add_account_to_list(
        self, index: int, account_id: str, account: msa.Account
    ) -> None:
        emoji = "✅" if accounts.is_active_account(account_id) else "👤"

        label = CTkLabel(
            master=self.accounts_list.content,
            text=f"{emoji} {account.minecraft_username}",
            anchor="w",
        )
        label.grid(row=index, column=0, pady=10, padx=0, sticky="nsw")
        remove_button = CTkButton(
            master=self.accounts_list.content,
            text="Remove 💣",
            width=0,
            command=lambda: self.remove_account(account_id),
        )
        remove_button.grid(row=index, column=1, pady=10, padx=10, sticky="nse")
        select_button = CTkButton(
            master=self.accounts_list.content,
            text="Select ✅",
            width=0,
            command=lambda: self.select_account(account_id),
        )
        select_button.grid(row=index, column=2, pady=10, padx=(0, 10), sticky="nse")

    def update_accounts_list(self) -> None:
        for account in self.accounts_list.content.winfo_children():
            account.destroy()

        for index, (account_id, account) in enumerate(accounts.get_accounts().items()):
            self.add_account_to_list(index * 2, account_id, account)

            separator = ttk.Separator(self.accounts_list.content, orient="horizontal")
            separator.grid(
                row=index * 2 + 1,
                column=0,
                columnspan=3,
                pady=0,
                padx=(0, 10),
                sticky="ew",
            )



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

import platform
import subprocess
from enum import Enum
from os import listdir, makedirs, path
from os import rename as mv
from shutil import rmtree
from subprocess import Popen
from threading import Thread
from typing import Callable

import tomli
import tomli_w
from pydantic import BaseModel

from . import (
    ASSETS_DIR,
    INSTANCES_DIR,
    ProgressCallbacks,
    __version__,
    accounts,
    jre_manager,
    minecraft_version_meta,
)
from .minecraft_rules import is_rule_list_valid
from .minecraft_version_meta import get_classpath_string
from .minecraft_versions import MinecraftVersionInfo, install_version

# flags from https://github.com/brucethemoose/Minecraft-Performance-Flags-Benchmarks
optimized_jvm_flags = [
    "-server",
    "-XX:+UseG1GC",
    "-XX:+ParallelRefProcEnabled",
    "-XX:MaxGCPauseMillis=50",
    "-XX:+UnlockExperimentalVMOptions",
    "-XX:+UnlockDiagnosticVMOptions",
    "-XX:+AlwaysPreTouch",
    "-XX:G1NewSizePercent=30",
    "-XX:G1MaxNewSizePercent=40",
    "-XX:G1HeapRegionSize=8M",
    "-XX:G1ReservePercent=20",
    "-XX:G1HeapWastePercent=5",
    "-XX:G1MixedGCCountTarget=4",
    "-XX:InitiatingHeapOccupancyPercent=15",
    "-XX:G1MixedGCLiveThresholdPercent=90",
    "-XX:G1RSetUpdatingPauseTimePercent=5",
    "-XX:SurvivorRatio=32",
    "-Dsun.rmi.dgc.server.gcInterval=2147483646",
    "-XX:+PerfDisableSharedMem",
    "-XX:MaxTenuringThreshold=1",
    "-XX:+UseStringDeduplication",
    "-XX:+UseFastUnorderedTimeStamps",
    "-XX:AllocatePrefetchStyle=1",
    "-XX:+OmitStackTraceInFastThrow",
    "-XX:ThreadPriorityPolicy=1",
    "-XX:+UseNUMA",
    "-XX:-DontCompileHugeMethods",
    "-XX:+UseVectorCmov",
    "-Djdk.nio.maxCachedBufferSize=262144",
    "-Dgraal.CompilerConfiguration=community",
    "-Dgraal.SpeculativeGuardMovement=true",
]


class InstanceType(str, Enum):
    vanilla = "vanilla"
    fabric = "fabric"
    forge = "forge"


class InstanceInfo(BaseModel):
    config_version: int = 1
    instance_type: InstanceType = InstanceType.vanilla
    minecraft_version: str = ""
    jre_version: str = "latest"


def list() -> list[str]:
    makedirs(INSTANCES_DIR, exist_ok=True)

    list = [
        instance_dir
        for instance_dir in listdir(INSTANCES_DIR)
        if path.isdir(path.join(INSTANCES_DIR, instance_dir))
    ]

    return list


def new(
    instance_name: str,
    minecraft_version: MinecraftVersionInfo,
    callbacks: ProgressCallbacks,
) -> None:
    print(f"Creating instance {instance_name}")

    instance_dir = path.join(INSTANCES_DIR, instance_name)
    makedirs(instance_dir)
    instance_info = InstanceInfo(minecraft_version=minecraft_version.id)
    _write_info(instance_name, instance_info)

    install_version(minecraft_version, callbacks)

    print("Done")


def _write_info(instance_name: str, instance_info: InstanceInfo) -> None:
    with open(path.join(INSTANCES_DIR, instance_name, "instance.toml"), "wb") as f:
        tomli_w.dump(instance_info.dict(), f)


def read_info(instance_name: str) -> InstanceInfo:
    with open(path.join(INSTANCES_DIR, instance_name, "instance.toml"), "rb") as f:
        info = InstanceInfo.parse_obj(tomli.load(f))

    # Writes the document file in case it is outdated.
    _write_info(instance_name, info)

    return info


def rename(old_name: str, new_name: str) -> None:
    old_dir = path.join(INSTANCES_DIR, old_name)
    new_dir = path.join(INSTANCES_DIR, new_name)
    mv(old_dir, new_dir)


def delete(instance_name: str) -> None:
    instance_dir = path.join(INSTANCES_DIR, instance_name)
    rmtree(instance_dir)


def launch(instance_name: str, account_id: str, callback_function: Callable) -> None:
    print(f"Launching instance {instance_name}")

    print("Refreshing account")
    account = accounts.refresh_account(account_id)
    print("Account successfully refreshed")

    instance_info = read_info(instance_name)
    version_meta = minecraft_version_meta.get_version_meta(
        instance_info.minecraft_version
    )

    jre_version = instance_info.jre_version
    if jre_version == "latest":
        jre_version = jre_manager.fetch_latest_java_version()

    is_updated = jre_manager.is_updated(jre_version)
    if not is_updated:
        print("Updating JRE")
        jre_manager.update(jre_version)

    java_path = jre_manager.get_java_path(jre_version)

    game_arguments = []
    for argument in version_meta.arguments.game:
        if isinstance(argument, minecraft_version_meta._ComplexArgument):
            if is_rule_list_valid(argument.rules):
                argument = argument.value
            else:
                argument = None

        if argument:
            match argument:
                case "${auth_player_name}":
                    argument = account.minecraft_username
                case "${version_name}":
                    argument = instance_info.minecraft_version
                case "${game_directory}":
                    argument = path.join(INSTANCES_DIR, instance_name)
                case "${assets_root}":
                    argument = ASSETS_DIR
                case "${assets_index_name}":
                    argument = version_meta.assetIndex.id
                case "${auth_uuid}":
                    argument = account_id
                case "${auth_access_token}":
                    argument = account.minecraft_access_token
                case "${clientid}":
                    argument = f"ice-launcher/{__version__}"
                case "${auth_xuid}":
                    argument = "0"
                case "${user_type}":
                    argument = "mojang"
                case "${version_type}":
                    argument = instance_info.instance_type.value

            game_arguments.append(argument)

    jvm_arguments = []
    if platform.system() == "Darwin":
        jvm_arguments.append("-XstartOnFirstThread")
    elif platform.system() == "Windows":
        jvm_arguments.append(
            "-XX:HeapDumpPath=MojangTricksIntelDriversForPerformance_javaw.exe_minecraft.exe.heapdump"
        )
        if platform.release() == "10":
            jvm_arguments.append("-Dos.name=Windows 10")
            jvm_arguments.append("-Dos.version=10.0")
    if platform.machine() in ["x86", "i386", "i686"]:
        jvm_arguments.append("-Xss1M")

    jvm_arguments.append("-Djava.library.path=natives")
    jvm_arguments.append("-Dminecraft.launcher.brand=ice-launcher")
    jvm_arguments.append(f"-Dminecraft.launcher.version={__version__}")
    jvm_arguments.append("-cp")
    jvm_arguments.append(
        get_classpath_string(version_meta.libraries, instance_info.minecraft_version)
    )

    command = [
        java_path,
        *jvm_arguments,
        *optimized_jvm_flags,
        version_meta.mainClass,
        *game_arguments,
    ]

    def start():
        process = Popen(
            command, cwd=path.join(INSTANCES_DIR, instance_name), stdout=subprocess.PIPE
        )
        callback_function(process)

    Thread(target=start).start()



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

import platform
from multiprocessing.pool import AsyncResult, ThreadPool
from os import path

from pydantic import BaseModel, HttpUrl

from . import LIBRARIES_DIR, VERSIONS_DIR, ProgressCallbacks, download_file
from .minecraft_assets import AssetIndex
from .minecraft_libraries import Library, get_valid_artifacts
from .minecraft_rules import Rule


class _ComplexArgument(BaseModel):
    rules: list[Rule]
    value: str | list[str]


class _Arguments(BaseModel):
    game: list[str | _ComplexArgument]


class _Artifact(BaseModel):
    sha1: str
    size: int
    url: HttpUrl


class _Downloads(BaseModel):
    client: _Artifact


class _JavaVersion(BaseModel):
    component: str


class MinecraftVersionMeta(BaseModel):
    arguments: _Arguments
    assetIndex: AssetIndex
    downloads: _Downloads
    javaVersion: _JavaVersion
    libraries: list[Library]
    mainClass: str


def get_version_meta(version_id: str) -> MinecraftVersionMeta:
    version_meta_path = path.join(VERSIONS_DIR, version_id, "meta.json")
    version_meta = MinecraftVersionMeta.parse_file(version_meta_path)

    return version_meta


def install_client(
    version_id: str, artifact: _Artifact, callbacks: ProgressCallbacks, pool: ThreadPool
) -> list[AsyncResult]:
    client_path = path.join(VERSIONS_DIR, version_id, "client.jar")

    result = pool.apply_async(
        download_file,
        (
            artifact.url,
            client_path,
            artifact.sha1,
            callbacks,
        ),
    )

    return [result]


def get_client_path(version_id: str) -> str:
    return path.join(VERSIONS_DIR, version_id, "client.jar")


def get_classpath_string(libraries: list[Library], minecraft_version: str) -> str:
    classpath_separator = ";" if platform.system() == "Windows" else ":"
    artifacts = get_valid_artifacts(libraries)

    jars = [path.join(LIBRARIES_DIR, artifact.path) for artifact in artifacts]
    jars.append(get_client_path(minecraft_version))

    classpath_string = classpath_separator.join(jars)

    return classpath_string



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

import platform
from os import listdir, makedirs, path, remove
from shutil import rmtree, unpack_archive
from typing import Final

import httpx
from packaging import version
from pydantic import BaseModel, HttpUrl

from . import JRES_DIR, download_file, headers

ADOPTIUM_API_ENDPOINT: Final[str] = "https://api.adoptium.net"


class _Package(BaseModel):
    checksum: str
    link: HttpUrl
    name: str
    size: int


class _Version(BaseModel):
    semver: str


class _Binary(BaseModel):
    package: _Package


class _Assets(BaseModel):
    binary: _Binary
    version: _Version


def fetch_latest_java_version() -> str:
    path = "/v3/info/available_releases"
    response = httpx.get(f"{ADOPTIUM_API_ENDPOINT}{path}", headers=headers)
    latest_release = response.json()["most_recent_feature_release"]

    return latest_release


def _get_architecture() -> str:
    if platform.machine() in ["AMD64", "x86_64"]:
        return "x64"
    if platform.machine() in ["i386", "i686", "x86"]:
        return "x86"
    if platform.machine() in ["aarch64", "arm64"]:
        return "aarch64"

    raise Exception("Unsupported architecture")


def _get_os() -> str:
    if platform.system() == "Linux":
        return "linux"
    if platform.system() == "Windows":
        return "windows"
    if platform.system() == "Darwin":
        return "mac"

    raise Exception("Unsupported OS")


def _get_assets_info(java_version: str) -> _Assets:
    url_path = f"/v3/assets/latest/{java_version}/hotspot"
    params = {
        "architecture": _get_architecture(),
        "image_type": "jre",
        "os": _get_os(),
        "vendor": "eclipse",
    }

    response = httpx.get(
        f"{ADOPTIUM_API_ENDPOINT}{url_path}",
        headers=headers | {"Accept": "application/json"},
        params=params,
    )
    assets_info_list = response.json()

    return _Assets.parse_obj(assets_info_list[0])


def is_updated(java_version: str) -> bool:
    makedirs(JRES_DIR, exist_ok=True)

    assets_info = _get_assets_info(java_version)

    current_semver = [
        dir for dir in listdir(JRES_DIR) if path.isdir(path.join(JRES_DIR, dir))
    ]
    if len(current_semver) == 0:
        return False

    current_semver = version.parse(
        current_semver[0].replace("jdk-", "").replace("-jre", "")
    )
    latest_semver = version.parse(assets_info.version.semver)

    return current_semver >= latest_semver


def update(java_version: str) -> None:
    makedirs(JRES_DIR, exist_ok=True)

    # To be deleted
    previous_files = listdir(JRES_DIR)

    assets_info = _get_assets_info(java_version)
    download_url = assets_info.binary.package.link

    extension = "zip" if platform.system() == "Windows" else "tar.gz"
    download_path = path.join(JRES_DIR, f"{assets_info.version.semver}.{extension}")

    download_file(
        url=download_url,
        dest=download_path,
    )

    unpack_archive(download_path, JRES_DIR)
    remove(download_path)

    # Delete previous files
    for file in previous_files:
        file_path = path.join(JRES_DIR, file)
        if path.isdir(file_path):
            rmtree(file_path)

        if path.isfile(file_path):
            remove(file_path)


def get_java_path(version: str) -> str:
    makedirs(JRES_DIR, exist_ok=True)

    available_jres = [
        dir for dir in listdir(JRES_DIR) if path.isdir(path.join(JRES_DIR, dir))
    ]

    current_jre = [dir for dir in available_jres if dir.startswith(f"jdk-{version}")][0]

    if platform.system() == "Windows":
        return path.join(JRES_DIR, current_jre, "bin", "java.exe")
    if platform.system() == "Darwin":
        return path.join(JRES_DIR, current_jre, "Contents", "Home", "bin", "java")
    if platform.system() == "Linux":
        return path.join(JRES_DIR, current_jre, "bin", "java")

    raise Exception("JRE not found")



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

from enum import Enum
from multiprocessing.pool import ThreadPool
from os import makedirs, path
from typing import Final

import httpx
from pydantic import BaseModel, HttpUrl

from . import VERSION_MANIFESTS_DIR, ProgressCallbacks, download_file, headers
from .minecraft_assets import get_total_assets_size, install_assets
from .minecraft_libraries import get_total_libraries_size, install_libraries
from .minecraft_version_meta import MinecraftVersionMeta, install_client

VERSION_MANIFEST_URL: Final[
    str
] = "https://piston-meta.mojang.com/mc/game/version_manifest_v2.json"


class _TypeEnum(str, Enum):
    release = "release"
    snapshot = "snapshot"
    old_beta = "old_beta"
    old_alpha = "old_alpha"


class MinecraftVersionInfo(BaseModel):
    id: str
    type: _TypeEnum
    url: HttpUrl
    sha1: str


class LatestVersions(BaseModel):
    release: str
    snapshot: str


class MinecraftVersionManifest(BaseModel):
    latest: LatestVersions
    versions: list[MinecraftVersionInfo]


def fetch_manifest() -> MinecraftVersionManifest:
    response = httpx.get(VERSION_MANIFEST_URL, headers=headers)
    manifest = MinecraftVersionManifest.parse_raw(response.content)

    return manifest


def install_version(
    minecraft_version: MinecraftVersionInfo, callbacks: ProgressCallbacks
) -> None:
    makedirs(VERSION_MANIFESTS_DIR, exist_ok=True)

    version_meta_path = path.join(VERSION_MANIFESTS_DIR, f"{minecraft_version.id}.json")
    download_file(
        url=minecraft_version.url,
        dest=version_meta_path,
        sha1hash=minecraft_version.sha1,
    )

    version_meta = MinecraftVersionMeta.parse_file(version_meta_path)

    total_size = (
        get_total_assets_size(version_meta.assetIndex)
        + get_total_libraries_size(version_meta.libraries)
        + version_meta.downloads.client.size
    )
    callbacks.set_status("Downloading required files")
    callbacks.set_max(total_size)

    with ThreadPool() as pool:
        results = install_assets(version_meta.assetIndex, callbacks, pool)
        results += install_libraries(version_meta.libraries, callbacks, pool)
        results += install_client(
            minecraft_version.id, version_meta.downloads.client, callbacks, pool
        )

        for result in results:
            result.wait()



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

from threading import Thread

from customtkinter import (
    CTkButton,
    CTkComboBox,
    CTkEntry,
    CTkFrame,
    CTkLabel,
    CTkProgressBar,
    StringVar,
)

from ice_launcher.components.heading import Heading
from ice_launcher.lib import ProgressCallbacks, instances, minecraft_versions

from .instances import Instances


class NewInstance(CTkFrame):
    def __init__(self, master) -> None:
        super().__init__(master=master)

        self.grid_columnconfigure(0, weight=1)

        heading = Heading(master=self, text="✨ Create a new instance")
        heading.grid(row=0, column=0, pady=20, padx=20, sticky="nwe")

        self.instance_name = CTkEntry(master=self, placeholder_text="My new instance")
        self.instance_name.grid(row=1, column=0, pady=10, padx=20, sticky="nswe")

        self.version = StringVar()
        self.version_selector = CTkComboBox(master=self, variable=self.version)
        self.version_selector.grid(row=2, column=0, pady=10, padx=20, sticky="nswe")

        # empty row as spacing
        self.grid_rowconfigure(3, weight=1)

        self.create_button = CTkButton(
            master=self,
            text="Create",
            command=self.create_instance,
        )
        self.create_button.grid(row=4, column=0, pady=(10, 20), padx=20, sticky="swe")

        Thread(target=self.update_versions).start()

    def update_versions(self) -> None:
        self.version_manifest = minecraft_versions.fetch_manifest()
        self.version_ids = [version.id for version in self.version_manifest.versions]
        self.version_selector.configure(values=self.version_ids)
        self.version.set(self.version_manifest.latest.release)

    def create_instance(self) -> None:
        instance_name = self.instance_name.get()
        version_id = self.version.get()

        for widget in self.winfo_children():
            widget.destroy()

        heading = Heading(master=self, text=f"🔄 Creating instance {instance_name}")
        heading.grid(row=0, column=0, pady=20, padx=20, sticky="nwe")

        self.grid_rowconfigure(1, weight=1)

        self.status_label = CTkLabel(
            master=self, text=f"Creating instance {instance_name}"
        )
        self.status_label.grid(row=2, column=0, pady=10, padx=20)

        self.progress_bar = CTkProgressBar(master=self)
        self.progress_bar.grid(row=3, column=0, pady=10, padx=20)

        self.grid_rowconfigure(4, weight=1)

        self.current_value = 0
        self.max_progress = 0

        def set_status(status: str) -> None:
            self.status_label.configure(text=status)

        def increment_value_by(value: int) -> None:
            self.current_value += value
            self.progress_bar.set(self.current_value / self.max_progress)

        def set_max(value: int) -> None:
            self.max_progress = value

        def reset() -> None:
            self.current_value = 0
            self.max_progress = 0
            self.progress_bar.set(0)

        callbacks = ProgressCallbacks(
            set_max=set_max,
            increment_value_by=increment_value_by,
            set_status=set_status,
            reset=reset,
        )

        def new_instance():
            version = next(
                version
                for version in self.version_manifest.versions
                if version.id == version_id
            )

            instances.new(
                instance_name,
                version,
                callbacks,
            )
            self.master.open_page(None, Instances(master=self.master))  # type: ignore

        Thread(target=new_instance).start()



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

from importlib import resources
from threading import Thread
from tkinter import PhotoImage
from typing import Optional

from customtkinter import (
    CTk,
    CTkButton,
    CTkFrame,
    set_appearance_mode,
    set_default_color_theme,
)

from ice_launcher.lib import launcher_config, launcher_updater
from ice_launcher.views import About, Accounts, Instances, Logs, News, Settings, Update

set_appearance_mode("dark")
set_default_color_theme("blue")


class App(CTk):
    current_active_button: Optional[CTkButton] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.title("Ice Launcher")
        self.geometry("780x520")
        self.protocol("WM_DELETE_WINDOW", self.on_closing)

        # set app icon
        with resources.path("ice_launcher.assets", "ice-launcher.png") as image_path:
            self.app_icon = PhotoImage(file=image_path)
            self.wm_iconphoto(True, self.app_icon)

        # configure grid layout (2x1)
        self.grid_columnconfigure(1, weight=1)
        self.grid_rowconfigure(0, weight=1)

        navigator = CTkFrame(master=self, width=180, corner_radius=0)  # type: ignore
        navigator.grid(row=0, column=0, sticky="nswe")

        instances_button = CTkButton(
            master=navigator,
            text="Instances",
            command=lambda: self.open_page(instances_button, Instances(master=self)),
            border_width=2,
            fg_color=None,
        )
        instances_button.grid(row=0, column=0, pady=(20, 10), padx=20, sticky="nswe")

        news_button = CTkButton(
            master=navigator,
            text="News",
            command=lambda: self.open_page(news_button, News(master=self)),
            border_width=2,
            fg_color=None,
        )
        news_button.grid(row=1, column=0, pady=10, padx=20)

        accounts_button = CTkButton(
            master=navigator,
            text="Accounts",
            command=lambda: self.open_page(accounts_button, Accounts(master=self)),
            border_width=2,
            fg_color=None,
        )
        accounts_button.grid(row=2, column=0, pady=10, padx=20)

        settings_button = CTkButton(
            master=navigator,
            text="Settings",
            command=lambda: self.open_page(settings_button, Settings(master=self)),
            border_width=2,
            fg_color=None,
        )
        settings_button.grid(row=3, column=0, pady=10, padx=20)

        # empty row as spacing
        navigator.grid_rowconfigure(50, weight=1)

        logs_button = CTkButton(
            master=navigator,
            text="Logs",
            command=lambda: self.open_page(logs_button, Logs(master=self)),
            border_width=2,
            fg_color=None,
        )
        logs_button.grid(row=100, column=0, pady=10, padx=20)

        about_button = CTkButton(
            master=navigator,
            text="About",
            command=lambda: self.open_page(about_button, About(master=self)),
            border_width=2,
            fg_color=None,
        )
        about_button.grid(row=101, column=0, pady=(10, 20), padx=20)

        self.current_page = CTkFrame(master=self)
        self.open_page(instances_button, Instances(master=self))

        if launcher_config.read().automatically_check_for_updates:
            Thread(target=self.check_for_updates).start()

    def on_closing(self, event=0) -> None:
        self.destroy()

    def open_page(self, button: Optional[CTkButton], page: CTkFrame) -> None:
        if self.current_active_button:
            self.current_active_button.configure(fg_color=None)

        if button:
            self.current_active_button = button
            self.current_active_button.configure(fg_color="#1F6AA5")

        for widget in self.current_page.winfo_children():
            widget.destroy()
        self.current_page.destroy()

        self.current_page = page
        self.current_page.grid(row=0, column=1, pady=20, padx=20, sticky="nswe")

    def check_for_updates(self) -> None:
        latest_version = launcher_updater.check_for_updates()
        if latest_version:
            Update(master=self, latest_version=latest_version)


def main():
    App().mainloop()


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

import webbrowser
from http.server import BaseHTTPRequestHandler, HTTPServer
from os import path
from threading import Thread
from typing import Any, Optional

import tomli
import tomli_w
from pydantic import BaseModel

from . import ACCOUNTS_FILE_PATH, msa


class Document(BaseModel):
    version: int = 1
    active_account: str = ""
    accounts: dict[str, msa.Account] = {}


def _write_document(doc: Document) -> None:
    with open(ACCOUNTS_FILE_PATH, "wb") as f:
        tomli_w.dump(doc.dict(), f)


def _read_document() -> Document:
    if not path.exists(ACCOUNTS_FILE_PATH):
        return Document()

    with open(ACCOUNTS_FILE_PATH, "rb") as f:
        doc = Document.parse_obj(tomli.load(f))

    # Writes the document file in case it is outdated.
    _write_document(doc)

    return doc


def get_accounts() -> dict[str, msa.Account]:
    doc = _read_document()

    return doc.accounts


class CallbackHandler(BaseHTTPRequestHandler):
    state: str
    code_verifier: str
    callback_function: Any

    def do_GET(self) -> None:
        account_entry = msa.login(self.path, self.state, self.code_verifier)
        add_account(account_entry)

        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(b"<html><head><title>Login</title></head>")
        self.wfile.write(b"<body><p>You have been logged in.</p>")
        self.wfile.write(b"<p>You can close this window.</p>")
        self.wfile.write(b"</body></html>")

        self.callback_function()
        Thread(target=self.server.shutdown).start()


def login_account(callback_function: Any) -> None:
    login_data = msa.get_login_data()

    webbrowser.open(login_data.uri)
    handler = CallbackHandler
    handler.state = login_data.state
    handler.code_verifier = login_data.code_verifier
    handler.callback_function = callback_function
    httpd = HTTPServer(("127.0.0.1", 3003), handler)
    httpd.serve_forever()


def add_account(account_entry: msa.AccountEntry) -> None:
    doc = _read_document()
    doc.accounts[account_entry.minecraft_id] = account_entry.account
    _write_document(doc)


def remove_account(account_id: str) -> None:
    doc = _read_document()

    del doc.accounts[account_id]
    if doc.active_account == account_id:
        doc.active_account = ""

    _write_document(doc)


def refresh_account(account_id: str) -> msa.Account:
    doc = _read_document()

    account = doc.accounts[account_id]
    account = msa.refresh(account)
    doc.accounts[account_id] = account

    _write_document(doc)

    return account


def get_active_account() -> Optional[msa.AccountEntry]:
    doc = _read_document()
    id = doc.active_account

    if id == "":
        return None

    account_entry = msa.AccountEntry(
        minecraft_id=id,
        account=doc.accounts[id],
    )

    return account_entry


def set_active_account(account_id: str) -> None:
    doc = _read_document()
    doc.active_account = account_id
    _write_document(doc)


def is_active_account(account_id: str) -> bool:
    doc = _read_document()

    return doc.active_account == account_id



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

from .about import About
from .accounts import Accounts
from .edit_instance import EditInstance
from .instances import Instances
from .logs import Logs
from .new_instance import NewInstance
from .news import News
from .settings import Settings
from .update import Update

__all__ = [
    "About",
    "Accounts",
    "EditInstance",
    "Instances",
    "Logs",
    "NewInstance",
    "News",
    "Settings",
    "Update",
]



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

from multiprocessing.pool import AsyncResult, ThreadPool
from os import makedirs, path
from pathlib import Path
from typing import Final

from pydantic import BaseModel, HttpUrl

from . import ASSETS_DIR, ProgressCallbacks, __version__, download_file

ASSETS_DOWNLOAD_ENDPOINT: Final[str] = "https://resources.download.minecraft.net"


class _ObjectInfo(BaseModel):
    hash: str
    size: int


class _Assets(BaseModel):
    objects: dict[str, _ObjectInfo]


class AssetIndex(BaseModel):
    id: str
    sha1: str
    size: int
    totalSize: int
    url: HttpUrl


def get_total_assets_size(asset_index: AssetIndex) -> int:
    return asset_index.size + asset_index.totalSize


def install_assets(
    asset_index: AssetIndex, callbacks: ProgressCallbacks, pool: ThreadPool
) -> list[AsyncResult]:
    makedirs(ASSETS_DIR, exist_ok=True)
    makedirs(path.join(ASSETS_DIR, "indexes"), exist_ok=True)
    makedirs(path.join(ASSETS_DIR, "objects"), exist_ok=True)

    asset_index_path = path.join(ASSETS_DIR, "indexes", f"{asset_index.id}.json")
    download_file(
        url=asset_index.url,
        dest=asset_index_path,
        sha1hash=asset_index.sha1,
        callbacks=callbacks,
    )
    assets = _Assets.parse_file(asset_index_path)

    results = []
    for asset_info in assets.objects.values():
        asset_path = path.join(
            ASSETS_DIR, "objects", asset_info.hash[:2], asset_info.hash
        )
        asset_url = (
            f"{ASSETS_DOWNLOAD_ENDPOINT}/{asset_info.hash[:2]}/{asset_info.hash}"
        )

        parent_dir = Path(asset_path).parent.absolute()
        makedirs(parent_dir, exist_ok=True)

        result = pool.apply_async(
            download_file,
            (
                asset_url,
                asset_path,
                asset_info.hash,
                callbacks,
            ),
        )
        results.append(result)

    return results



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

from subprocess import Popen
from tkinter import ttk

from customtkinter import CTkButton, CTkFrame, CTkLabel

from ice_launcher import views
from ice_launcher.components.heading import Heading
from ice_launcher.components.scrollable_frame import ScrollableFrame
from ice_launcher.lib import accounts, instances


class Instances(CTkFrame):
    def __init__(self, master) -> None:
        super().__init__(master=master)

        self.account_list = list(accounts.get_accounts().keys())

        self.grid_columnconfigure(0, weight=1)

        heading = Heading(master=self, text="🧊 Instances")
        heading.grid(row=0, column=0, pady=20, padx=20, sticky="nwe")

        self.instances_list = ScrollableFrame(master=self)
        self.instances_list.grid(row=1, column=0, pady=20, padx=20, sticky="nswe")
        self.grid_rowconfigure(1, weight=1)

        status_bar = CTkFrame(master=self)
        status_bar.grid(row=100, column=0, pady=0, padx=0, sticky="swe")

        new_instance_button = CTkButton(
            master=status_bar,
            text="New Instance ✨",
            command=self.add_new_instance,
        )
        new_instance_button.grid(row=0, column=0, pady=10, padx=10, sticky="nsw")

        # empty column as spacing
        status_bar.grid_columnconfigure(1, weight=1)

        self.account_label = CTkLabel(
            master=status_bar,
            text="",
            anchor="e",
        )
        self.account_label.grid(row=0, column=2, pady=0, padx=10, sticky="nse")

        self.update_selected_account()
        self.update_instance_list()

    def add_new_instance(self) -> None:
        self.master.open_page(None, views.NewInstance(master=self.master))  # type: ignore

    def edit_instance(self, instance_name: str) -> None:
        self.master.open_page(None, views.EditInstance(master=self.master, instance_name=instance_name))  # type: ignore

    def launch_instance(self, instance_name: str) -> None:
        self.master.open_page(None, views.Logs(master=self.master))  # type: ignore

        def show_logs(process: Popen):
            self.master.current_page.update_logs(process)  # type: ignore

        if self.selected_account is not None:
            instances.launch(
                instance_name, self.selected_account.minecraft_id, show_logs
            )

    def update_selected_account(self) -> None:
        self.selected_account = accounts.get_active_account()

        if self.selected_account is None:
            account_name = "You need to select an account first"
        else:
            account_name = (
                f"Account: {self.selected_account.account.minecraft_username}"
            )

        self.account_label.configure(text=account_name)

    def add_instance_to_list(self, index: int, instance_name: str) -> None:
        match instances.read_info(instance_name).instance_type:
            case instances.InstanceType.vanilla:
                instance_emoji = "🍦"
            case instances.InstanceType.fabric:
                instance_emoji = "🧵"
            case instances.InstanceType.forge:
                instance_emoji = "⚒️"

        label = CTkLabel(
            master=self.instances_list.content,
            text=f"{instance_emoji} {instance_name}",
            anchor="w",
        )
        label.grid(row=index, column=0, pady=10, padx=0, sticky="nsw")
        edit_button = CTkButton(
            master=self.instances_list.content,
            text="Edit ⚙️",
            width=0,
            command=lambda: self.edit_instance(instance_name),
        )
        edit_button.grid(row=index, column=1, pady=10, padx=0, sticky="nse")

        if self.selected_account is not None:
            launch_button = CTkButton(
                master=self.instances_list.content,
                text="Launch 🚀",
                width=0,
                command=lambda: self.launch_instance(instance_name),
            )
            launch_button.grid(row=index, column=2, pady=10, padx=10, sticky="nse")

    def update_instance_list(self) -> None:
        instance_list = instances.list()

        for instance in self.instances_list.content.winfo_children():
            instance.destroy()

        for index, instance_name in enumerate(instance_list):
            self.add_instance_to_list(index * 2, instance_name)

            separator = ttk.Separator(self.instances_list.content, orient="horizontal")
            separator.grid(
                row=index * 2 + 1,
                column=0,
                columnspan=3,
                pady=0,
                padx=(0, 10),
                sticky="ew",
            )



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

from tkinter import ttk

from customtkinter import (
    BooleanVar,
    CTkButton,
    CTkEntry,
    CTkFrame,
    CTkLabel,
    CTkSwitch,
    StringVar,
)

from ice_launcher.components.heading import Heading
from ice_launcher.components.scrollable_frame import ScrollableFrame
from ice_launcher.lib import launcher_config


class Settings(CTkFrame):
    def __init__(self, master) -> None:
        super().__init__(master=master)
        self.grid_columnconfigure(0, weight=1)

        heading = Heading(master=self, text="⚙️ Settings")
        heading.grid(row=0, column=0, pady=20, padx=20, sticky="nwe")

        settings_frame = ScrollableFrame(master=self)
        settings_frame.grid(row=1, column=0, pady=20, padx=20, sticky="nswe")
        self.grid_rowconfigure(1, weight=1)

        self.automatically_check_for_updates = BooleanVar()
        automatically_check_for_updates_label = CTkLabel(
            master=settings_frame.content,
            text="Automatically check for updates",
            anchor="w",
        )
        automatically_check_for_updates_label.grid(
            row=0, column=0, pady=10, padx=0, sticky="w"
        )
        automatically_check_for_updates_switch = CTkSwitch(
            master=settings_frame.content,
            variable=self.automatically_check_for_updates,
            text="",
        )
        automatically_check_for_updates_switch.grid(
            row=0, column=1, pady=10, padx=(0, 10), sticky="e"
        )

        separator = ttk.Separator(settings_frame.content, orient="horizontal")
        separator.grid(row=1, column=0, columnspan=2, pady=0, padx=(0, 10), sticky="ew")

        self.jvm_arguments = StringVar()
        jvm_arguments_label = CTkLabel(
            master=settings_frame.content, text="JVM arguments", anchor="w"
        )
        jvm_arguments_label.grid(row=2, column=0, pady=10, padx=0, sticky="w")
        jvm_arguments_entry = CTkEntry(
            master=settings_frame.content, textvariable=self.jvm_arguments
        )
        jvm_arguments_entry.grid(row=2, column=1, pady=10, padx=(0, 10), sticky="e")

        separator = ttk.Separator(settings_frame.content, orient="horizontal")
        separator.grid(row=3, column=0, columnspan=2, pady=0, padx=(0, 10), sticky="ew")

        self.jvm_memory = StringVar()
        jvm_memory_label = CTkLabel(
            master=settings_frame.content, text="JVM memory", anchor="w"
        )
        jvm_memory_label.grid(row=4, column=0, pady=10, padx=0, sticky="w")
        jvm_memory_entry = CTkEntry(
            master=settings_frame.content, textvariable=self.jvm_memory
        )
        jvm_memory_entry.grid(row=4, column=1, pady=10, padx=(0, 10), sticky="e")

        self.button_bar = CTkFrame(master=self)
        self.button_bar.grid(row=100, column=0, pady=0, padx=0, sticky="swe")

        # empty column as spacing
        self.button_bar.grid_columnconfigure(0, weight=1)

        self.reset_button = CTkButton(
            master=self.button_bar,
            text="Reset to default settings ♻️",
            fg_color=None,
            border_width=2,
            command=self.reset_to_default_settings,
        )
        self.reset_button.grid(row=0, column=1, pady=10, padx=10, sticky="nse")

        self.save_button = CTkButton(
            master=self.button_bar,
            text="Save ✅",
            command=self.save,
        )
        self.save_button.grid(row=0, column=2, pady=10, padx=10, sticky="nse")

        self.update_settings(launcher_config.read())

    def update_settings(self, config: launcher_config.Config) -> None:
        self.automatically_check_for_updates.set(config.automatically_check_for_updates)
        self.jvm_arguments.set(" ".join(config.jvm_arguments))
        self.jvm_memory.set(config.jvm_memory)

    def reset_to_default_settings(self) -> None:
        default_config = launcher_config.Config()
        self.update_settings(default_config)

    def save(self) -> None:
        config = launcher_config.read()

        config.automatically_check_for_updates = (
            self.automatically_check_for_updates.get()
        )
        config.jvm_arguments = self.jvm_arguments.get().split()
        config.jvm_memory = self.jvm_memory.get()

        launcher_config.write(config)



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

from tkinter import messagebox

from customtkinter import CTkButton, CTkFrame, CTkInputDialog

from ice_launcher import views
from ice_launcher.components.heading import Heading
from ice_launcher.lib import instances


class EditInstance(CTkFrame):
    def __init__(self, master, instance_name: str) -> None:
        super().__init__(master=master)
        self.grid_columnconfigure(0, weight=1)

        self.heading = Heading(master=self, text=instance_name)
        self.heading.grid(row=0, column=0, pady=20, padx=20, sticky="nwe")

        # empty row as spacing
        self.grid_rowconfigure(99, weight=1)

        button_bar = CTkFrame(master=self)
        button_bar.grid(row=100, column=0, pady=20, padx=20, sticky="swe")

        # empty column as spacing
        button_bar.grid_columnconfigure(0, weight=1)

        rename_button = CTkButton(
            master=button_bar,
            text="Rename 📝",
            command=self.rename_instance,
        )
        rename_button.grid(row=0, column=1, pady=10, padx=10, sticky="nse")

        delete_button = CTkButton(
            master=button_bar,
            text="Delete 💣",
            command=self.delete_instance,
        )
        delete_button.grid(row=0, column=2, pady=10, padx=10, sticky="nse")

    def rename_instance(self):
        dialog = CTkInputDialog(
            master=None,
            text="Type a new name for the instance",
            title="Rename Instance",
        )
        new_name = dialog.get_input()
        if new_name:
            instances.rename(self.heading.label.text, new_name)
            self.heading.label.configure(text=new_name)

    def done(self):
        self.master.open_page(None, views.Instances(master=self.master))  # type: ignore

    def delete_instance(self) -> None:
        if messagebox.askyesno(
            "Delete Instance",
            f"Are you sure you want to delete the instance {self.heading.label.text}?",
        ):
            instances.delete(self.heading.label.text)
            self.master.open_page(None, views.Instances(master=self.master))  # type: ignore



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

from enum import Enum
from typing import Final, Optional

import httpx
from pydantic import BaseModel

from . import headers

MINECRAFT_NEWS_URL: Final[
    str
] = "https://www.minecraft.net/content/minecraft-net/_jcr_content.articles.grid"

http_client = httpx.Client(headers=headers)


class _ArticleLang(str, Enum):
    en_us = "en-us"


class _BackgroundColor(str, Enum):
    bg_blue = "bg-blue"
    bg_green = "bg-green"
    bg_red = "bg-red"


class _ContentType(str, Enum):
    image = "image"
    outgoing_link = "outgoing-link"
    video = "video"


class _Image(BaseModel):
    content_type: _ContentType
    imageURL: str
    alt: Optional[str] = None
    videoURL: Optional[str] = None
    videoType: Optional[str] = None
    videoProvider: Optional[str] = None
    videoId: Optional[str] = None
    linkurl: Optional[str] = None
    background_color: Optional[_BackgroundColor] = None


class _TileSize(str, Enum):
    the1x1 = "1x1"
    the1x2 = "1x2"
    the2x1 = "2x1"
    the2x2 = "2x2"
    the4x2 = "4x2"


class _Tile(BaseModel):
    sub_header: str
    image: _Image
    tile_size: _TileSize
    title: str


class _Article(BaseModel):
    default_tile: _Tile
    articleLang: _ArticleLang
    primary_category: str
    categories: list[str]
    article_url: str
    publish_date: str
    tags: list[str]
    preferred_tile: Optional[_Tile] = None


class Articles(BaseModel):
    article_grid: list[_Article]
    article_count: int


def fetch(page_size: int = 20) -> Articles:
    """
    Get the news from minecraft.net
    """

    parameters = {"pageSize": page_size}
    response = http_client.get(MINECRAFT_NEWS_URL, params=parameters)
    articles = Articles.parse_raw(response.content)

    return articles



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

import platform
from enum import Enum
from typing import Optional

from pydantic import BaseModel


class _Action(str, Enum):
    allow = "allow"


class _OsName(str, Enum):
    windows = "windows"
    linux = "linux"
    osx = "osx"


class _Os(BaseModel):
    name: _OsName


class _Feature(str, Enum):
    is_demo_user = "is_demo_user"
    has_custom_resolution = "has_custom_resolution"


class Rule(BaseModel):
    action: _Action
    os: Optional[_Os]
    features: Optional[dict[_Feature, bool]]


def is_rule_list_valid(rules: list[Rule]) -> bool:
    os_name = platform.system().lower().replace("darwin", "osx")

    for rule in rules:
        if rule.features:
            return False
        if rule.os:
            if rule.action == _Action.allow and rule.os.name == os_name:
                return True

    return False



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Final, Optional

import httpx
import packaging.version
from pydantic import BaseModel

from . import __version__, headers

LATEST_RELEASE_URL: Final[
    str
] = "https://api.github.com/repos/mq1/ice-launcher/releases/latest"


class _ReleaseInfo(BaseModel):
    tag_name: str


def _get_latest_release() -> str:
    """Get the latest release of the launcher."""

    response = httpx.get(LATEST_RELEASE_URL, headers=headers)
    release_info = _ReleaseInfo.parse_raw(response.content)

    return release_info.tag_name


def check_for_updates() -> Optional[str]:
    """
    Check if there is a new version of the launcher available.
    Returns the version number if there is a new version, None otherwise.
    """

    current_release = packaging.version.parse(__version__)
    latest_release = packaging.version.parse(_get_latest_release())

    return latest_release.__str__() if latest_release > current_release else None



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

from os import path

import tomli
import tomli_w
from pydantic import BaseModel

from . import CONFIG_PATH


class Config(BaseModel):
    config_version: int = 1
    automatically_check_for_updates: bool = True
    jvm_arguments: list[str] = []
    jvm_memory: str = "2G"


def write(config: Config) -> None:
    with open(CONFIG_PATH, "wb") as f:
        tomli_w.dump(config.dict(), f)


def read() -> Config:
    if not path.exists(CONFIG_PATH):
        return Config()

    with open(CONFIG_PATH, "rb") as f:
        config = Config.parse_obj(tomli.load(f))

    # Writes the config file in case it is outdated.
    write(config)

    return config



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only


import hashlib
import lzma
import os
import stat
from os import chmod, makedirs, path, remove
from typing import Callable, Final, Optional

import httpx
from appdirs import AppDirs
from pydantic import BaseModel

dirs: AppDirs = AppDirs("ice-launcher", "mq1.eu")
makedirs(dirs.user_data_dir, exist_ok=True)


__version__: Final[str] = "0.0.17"
CLIENT_ID: Final[str] = "0018ddff-bd2f-4cc6-b220-66f6a4462a5c"
VERSIONS_DIR: Final[str] = path.join(dirs.user_data_dir, "versions")
LIBRARIES_DIR: Final[str] = path.join(dirs.user_data_dir, "libraries")
ASSETS_DIR: Final[str] = path.join(dirs.user_data_dir, "assets")
JRES_DIR: Final[str] = path.join(dirs.user_data_dir, "jre")
CONFIG_PATH: Final[str] = path.join(dirs.user_data_dir, "config.toml")
VERSION_MANIFESTS_DIR: Final[str] = path.join(dirs.user_data_dir, "versions")
ACCOUNTS_FILE_PATH: Final[str] = path.join(dirs.user_data_dir, "accounts.toml")
INSTANCES_DIR: Final[str] = path.join(dirs.user_data_dir, "instances")


headers = {
    "user-agent": f"ice-launcher/{__version__}",
}


class ProgressCallbacks(BaseModel):
    set_max: Callable[[int], None]
    increment_value_by: Callable[[int], None]
    set_status: Callable[[str], None]
    reset: Callable[[], None]


def download_file(
    url: str,
    dest: str,
    sha1hash: Optional[str] = None,
    callbacks: Optional[ProgressCallbacks] = None,
    is_lzma: bool = False,
    set_executable: bool = False,
) -> None:
    # If the file already exists, we check if the hash matches.
    if path.exists(dest):
        print(f"File {dest} already exists, checking hash...")

        if sha1hash:
            sha1 = hashlib.sha1()

            with open(dest, "rb") as f:
                while chunk := f.read(65536):  # Read 64kb chunks.
                    sha1.update(chunk)

            if sha1.hexdigest() == sha1hash:
                if callbacks:
                    file_size = path.getsize(dest)
                    callbacks.increment_value_by(file_size)

                print(f"{dest} Hash matches, skipping download.")
                return

        print(f"{dest} Hash does not match, redownloading.")
        remove(dest)

    print("Downloading file from", url, "to", dest)

    with open(dest, "wb") as file:
        with httpx.stream("GET", url, headers=headers) as response:
            for chunk in response.iter_bytes(65536):  # Read 64kb chunks.
                if callbacks:
                    callbacks.increment_value_by(len(chunk))

                if is_lzma:
                    chunk = lzma.decompress(chunk)

                file.write(chunk)

    print(f"Downloaded {dest}")

    if set_executable:
        st = os.stat(dest)
        chmod(dest, st.st_mode | stat.S_IEXEC)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ice_launcher',
 'ice_launcher.assets',
 'ice_launcher.components',
 'ice_launcher.lib',
 'ice_launcher.views']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4,<2.0.0',
 'customtkinter>=4.5.10,<5.0.0',
 'httpx>=0.23.0,<0.24.0',
 'packaging>=21.3,<22.0',
 'pydantic>=1.9.2,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'tomli-w>=1.0.0,<2.0.0',
 'tomli>=2.0.1,<3.0.0']

entry_points = \
{'console_scripts': ['start = ice_launcher.__main__:main']}

setup_kwargs = {
    'name': 'ice-launcher',
    'version': '0.0.17',
    'description': 'An unofficial WIP Minecraft Launcher',
    'long_description': '<p align="center">\n<img src="./ice_launcher/assets/ice-launcher.png" alt="Ice Launcher Logo" height="128px">\n</p>\n\n<h1 align="center">Ice Launcher</h1>\n\n<p align="center">\n<a href="https://github.com/mq1/ice-launcher/blob/main/LICENSE"><img alt="License: GPL-3.0" src="https://img.shields.io/github/license/mq1/ice-launcher"></a>\n<a href="https://github.com/psf/black"><img alt="Code style: black" src="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n</p>\n\n-----\n\nAn unofficial WIP launcher for Minecraft 1.19+\n\nAs of now, this launcher is considered a technical preview and every release breaks everything from the previous install\n\n**Table of Contents**\n\n- [Installation](#installation)\n- [Thanks](#thanks)\n- [License](#license)\n\n## Installation\n\n### Package (recommended)\n\nJust grab the [latest release](https://github.com/mq1/ice-launcher/releases/latest) for your platform\n\n### From pip\n\n```sh\npip install ice-launcher\n```\n\n## Thanks\n\n- Aru for the awesome logo!\n- [JakobDev](https://gitlab.com/JakobDev) for building [minecraft-launcher-lib](https://gitlab.com/JakobDev/minecraft-launcher-lib)\n- [TomSchimansky](https://github.com/TomSchimansky) for building [CustomTkinter](https://github.com/TomSchimansky/CustomTkinter)\n\n## License\n\n`ice-launcher` is distributed under the terms of the [GPL-3.0-only](https://spdx.org/licenses/GPL-3.0-only.html) license.\n',
    'author': 'Manuel Quarneti',
    'author_email': 'manuelquarneti@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mq1/ice-launcher#readme',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<3.11',
}


setup(**setup_kwargs)



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

import platform
from multiprocessing.pool import AsyncResult, ThreadPool
from os import makedirs, path
from pathlib import Path
from typing import Optional

from pydantic import BaseModel, HttpUrl

from . import LIBRARIES_DIR, ProgressCallbacks, download_file
from .minecraft_rules import Rule, is_rule_list_valid


class _Artifact(BaseModel):
    path: str
    sha1: str
    size: int
    url: HttpUrl


class _Downloads(BaseModel):
    artifact: _Artifact
    rules: Optional[list[Rule]]


class Library(BaseModel):
    downloads: _Downloads


def get_natives_string() -> str:
    """
    Returns the natives string for the current system
    """
    match platform.system():
        case "Linux":
            return "natives-linux"
        case "Darwin":
            match platform.machine():
                case "x86_64":
                    return "natives-macos"
                case "arm64":
                    return "natives-macos-arm64"
        case "Windows":
            match platform.machine():
                case "AMD64":
                    return "natives-windows"
                case "x86":
                    return "natives-windows-x86"

    raise Exception("Unsupported platform")


def get_total_libraries_size(libraries: list[Library]) -> int:
    return sum(library.downloads.artifact.size for library in libraries)


def get_valid_artifacts(libraries: list[Library]) -> list[_Artifact]:
    natives_string = get_natives_string()

    valid_artifacts = []
    for library in libraries:
        if library.downloads.rules and not is_rule_list_valid(library.downloads.rules):
            continue

        if "natives" in library.downloads.artifact.path:
            if natives_string not in library.downloads.artifact.path:
                continue

        if (
            "x86_64" in library.downloads.artifact.path
            and platform.machine() != "x86_64"
        ):
            continue
        if (
            "aarch_64" in library.downloads.artifact.path
            and platform.machine() != "aarch64"
        ):
            continue

        valid_artifacts.append(library.downloads.artifact)

    return valid_artifacts


def install_libraries(
    libraries: list[Library], callbacks: ProgressCallbacks, pool: ThreadPool
) -> list[AsyncResult]:
    artifacts = get_valid_artifacts(libraries)

    results = []
    for artifact in artifacts:
        library_path = path.join(LIBRARIES_DIR, artifact.path)
        parent_dir = Path(library_path).parent.absolute()
        makedirs(parent_dir, exist_ok=True)

        result = pool.apply_async(
            download_file,
            (
                artifact.url,
                library_path,
                artifact.sha1,
                callbacks,
            ),
        )
        results.append(result)

    return results



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

from subprocess import Popen
from tkinter import END, Text

from customtkinter import CTkFrame

from ice_launcher.components.heading import Heading


class Logs(CTkFrame):
    def __init__(self, master) -> None:
        super().__init__(master=master)
        self.grid_columnconfigure(0, weight=1)

        heading = Heading(master=self, text="🗒 Logs")
        heading.grid(row=0, column=0, columnspan=2, pady=20, padx=20, sticky="nwe")

        self.logs = Text(master=self, highlightthickness=0)
        self.logs.grid(row=1, column=0, pady=20, padx=20, sticky="nswe")
        self.grid_rowconfigure(1, weight=1)

    def update_logs(self, process: Popen):
        while process.poll() is None:
            output = process.stdout.readline()  # type: ignore
            self.logs.insert(END, output)



# SPDX-FileCopyrightText: 2022-present Manuel Quarneti <manuelquarneti@gmail.com>
#
# SPDX-License-Identifier: GPL-3.0-only

import secrets
from base64 import urlsafe_b64encode
from enum import Enum
from hashlib import sha256
from typing import Final
from urllib.parse import parse_qs, urlencode, urlparse

import httpx
from pydantic import BaseModel

from . import CLIENT_ID, headers

MSA_AUTHORIZATION_ENDPOINT: Final[
    str
] = "https://login.microsoftonline.com/consumers/oauth2/v2.0/authorize"
MSA_TOKEN_ENDPOINT: Final[
    str
] = "https://login.microsoftonline.com/consumers/oauth2/v2.0/token"
XBOXLIVE_AUTH_ENDPOINT: Final[str] = "https://user.auth.xboxlive.com/user/authenticate"
XSTS_AUTHORIZATION_ENDPOINT: Final[
    str
] = "https://xsts.auth.xboxlive.com/xsts/authorize"
MINECRAFT_AUTH_ENDPOINT: Final[
    str
] = "https://api.minecraftservices.com/authentication/login_with_xbox"
MINECRAFT_PROFILE_ENDPOINT: Final[
    str
] = "https://api.minecraftservices.com/minecraft/profile"
SCOPE: Final[str] = "XboxLive.signin offline_access"
REDIRECT_URI: Final[str] = "http://localhost:3003"


class _CodeChallengeMethod(str, Enum):
    plain = "plain"
    S256 = "S256"


class _PKCE_Data(BaseModel):
    code_verifier: str
    code_challenge: str
    code_challenge_method: _CodeChallengeMethod


class _LoginData(BaseModel):
    uri: str
    state: str
    code_verifier: str


class _OAuth2Token(BaseModel):
    access_token: str
    refresh_token: str


class _Xui(BaseModel):
    uhs: str


class _DisplayClaims(BaseModel):
    xui: list[_Xui]


class _XBLResponse(BaseModel):
    Token: str
    DisplayClaims: _DisplayClaims


class _XSTSResponse(BaseModel):
    Token: str


class _MinecraftResponse(BaseModel):
    access_token: str


class _MinecraftProfile(BaseModel):
    id: str
    name: str


class Account(BaseModel):
    microsoft_refresh_token: str
    minecraft_access_token: str
    minecraft_username: str


class AccountEntry(BaseModel):
    minecraft_id: str
    account: Account


def _generate_pkce_data() -> _PKCE_Data:
    """
    Generates the PKCE code challenge and code verifier
    """
    code_verifier = secrets.token_urlsafe(128)[:128]
    code_challenge = urlsafe_b64encode(
        sha256(code_verifier.encode("ascii")).digest()
    ).decode("ascii")[:-1]
    code_challenge_method = _CodeChallengeMethod.S256

    return _PKCE_Data(
        code_verifier=code_verifier,
        code_challenge=code_challenge,
        code_challenge_method=code_challenge_method,
    )


def _generate_state() -> str:
    """
    Generates a random state
    """
    return secrets.token_urlsafe(16)


def get_login_data() -> _LoginData:
    pkce_data = _generate_pkce_data()
    state = _generate_state()

    data = {
        "client_id": CLIENT_ID,
        "response_type": "code",
        "redirect_uri": REDIRECT_URI,
        "response_mode": "query",
        "scope": SCOPE,
        "state": state,
        "code_challenge": pkce_data.code_challenge,
        "code_challenge_method": pkce_data.code_challenge_method.value,
    }

    uri = urlparse(MSA_AUTHORIZATION_ENDPOINT)._replace(query=urlencode(data)).geturl()

    return _LoginData(uri=uri, state=state, code_verifier=pkce_data.code_verifier)


def _get_minecraft_account_data(msa_token: _OAuth2Token) -> AccountEntry:
    http_client = httpx.Client(
        headers=headers
        | {"Content-Type": "application/json", "Accept": "application/json"}
    )

    # Authenticate with Xbox Live
    data = {
        "Properties": {
            "AuthMethod": "RPS",
            "SiteName": "user.auth.xboxlive.com",
            "RpsTicket": f"d={msa_token.access_token}",
        },
        "RelyingParty": "http://auth.xboxlive.com",
        "TokenType": "JWT",
    }
    response = http_client.post(XBOXLIVE_AUTH_ENDPOINT, json=data)
    xbl_response = _XBLResponse.parse_raw(response.content)

    # Authenticate with XSTS
    data = {
        "Properties": {"SandboxId": "RETAIL", "UserTokens": [xbl_response.Token]},
        "RelyingParty": "rp://api.minecraftservices.com/",
        "TokenType": "JWT",
    }
    response = http_client.post(XSTS_AUTHORIZATION_ENDPOINT, json=data)
    xsts_response = _XSTSResponse.parse_raw(response.content)

    # Authenticate with Minecraft
    data = {
        "identityToken": f"XBL3.0 x={xbl_response.DisplayClaims.xui[0].uhs};{xsts_response.Token}"
    }
    response = http_client.post(MINECRAFT_AUTH_ENDPOINT, json=data)
    minecraft_response = _MinecraftResponse.parse_raw(response.content)

    # Get Minecraft profile
    http_client.headers["Authorization"] = f"Bearer {minecraft_response.access_token}"
    response = http_client.get(MINECRAFT_PROFILE_ENDPOINT)
    minecraft_profile = _MinecraftProfile.parse_raw(response.content)

    account = Account(
        microsoft_refresh_token=msa_token.refresh_token,
        minecraft_access_token=minecraft_response.access_token,
        minecraft_username=minecraft_profile.name,
    )

    return AccountEntry(minecraft_id=minecraft_profile.id, account=account)


def login(authorization_response: str, state: str, code_verifier: str) -> AccountEntry:
    http_client = httpx.Client(
        headers=headers
        | {
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "application/json",
        }
    )

    url = urlparse(authorization_response)
    query = parse_qs(url.query)
    assert query["state"][0] == state

    data = {
        "client_id": CLIENT_ID,
        "scope": SCOPE,
        "code": query["code"][0],
        "redirect_uri": REDIRECT_URI,
        "grant_type": "authorization_code",
        "code_verifier": code_verifier,
    }
    response = http_client.post(MSA_TOKEN_ENDPOINT, data=data)
    msa_token = _OAuth2Token.parse_raw(response.content)
    account_entry = _get_minecraft_account_data(msa_token)

    return account_entry


def refresh(account: Account) -> Account:
    http_client = httpx.Client(
        headers=headers
        | {
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "application/json",
        }
    )

    data = {
        "client_id": CLIENT_ID,
        "scope": SCOPE,
        "refresh_token": account.microsoft_refresh_token,
        "grant_type": "refresh_token",
    }

    response = http_client.post(MSA_TOKEN_ENDPOINT, data=data)
    msa_token = _OAuth2Token.parse_raw(response.content)
    account_entry = _get_minecraft_account_data(msa_token)

    return account_entry.account
